/**
 * XCEBikeDevice.js
 * 设备信息及操作类
 * 只有操作方法，无操作策略。操作策略在service层维护
 * lyoncat
 */

const Sequelize = require('sequelize');
const moment = require('moment');
const XCYuntuUtil = require('./XCYuntuUtil');

let xcSequelize = null;
let xcRedis = null;
let xcYuntu = null;

/**
 * 设备状态表
 */
const XC_EBIKE_DEVICE_STATE = {
    READY: '1', //可使用
    RIDING: '2', //骑行中
    BROKEN: '3', //设备报修
    BOOKING: '4', //设备被预约
    OFFLINE: '5', //设备离线
    CHANGING_BATTERY: '6', //处于工人换电池阶段
};

const CREATE_RULE = {
    freezeTableName: true
}

//设备行程表结构
const XC_EBIKE_DEVICE_ITINERARY_TABLE_DS = {
    itinId: {
        type: Sequelize.INTEGER(16).UNSIGNED,
        primaryKey: true,
        autoIncrement: true
    },
    startTime: {
        type: Sequelize.DATE,
        unique: true,
        allowNull: false
    },
    startLat: {
        type: Sequelize.DOUBLE(9, 6),
    },
    startLng: {
        type: Sequelize.DOUBLE(9, 6),
    },
    endTime: {
        type: Sequelize.DATE,
        allowNull: false,
        unique: true,
    },
    endLat: {
        type: Sequelize.DOUBLE(9, 6),
    },
    endLng: {
        type: Sequelize.DOUBLE(9, 6),
    },
    itinerary: {
        type: Sequelize.INTEGER(8).UNSIGNED
    },

}

/**
 * 设备每次预约保留的时间
 */
let EBikeBookingTime = 5 * 60; //单位秒
// let EBikeBookingTime = 5; //单位秒

class XCEBikeDevice {
    static get XC_EBIKE_DEVICE_STATE() {
        return XC_EBIKE_DEVICE_STATE;
    }
    /**
     * 配置redis client
     * 和RedisSubClient，用于监听Key过期事件
     */
    static configRedis(redisClient, redisSubClient) {
        xcRedis = redisClient;
        XCEBikeDevice.configExpiredKeyListen(redisSubClient);
    }

    /**
     * 配置预约过期key事件监听
     */
    static configExpiredKeyListen(redisSubClient){
        if(xcRedis && redisSubClient){
            redisSubClient.subscribe(`__keyevent@${xcRedis.options.db}__:expired`, 1);//同一db
            redisSubClient.on('message', (msg, key) => {
                XCEBikeDevice.handleExpiredKeyEvents(msg, key);
              })
        }
        
    }

    /**
     * 处理过期消息和过期Key
     */
    static handleExpiredKeyEvents(msg, key){
        XCEBikeDevice.handleDeviceBookingExpired(msg, key);
    }

    /**
     * 配置数据库实例
     */
    static configSequelize(sequelizeInstance) {
        xcSequelize = sequelizeInstance;
    }

    /**
     * 配置yuntu实例
     */
    static async configYuntu(yuntu) {
        xcYuntu = yuntu;
        await XCYuntuUtil.deviceTableidEnsureInit(yuntu);
    }

    static log() {
        console.log(xcRedis, xcSequelize, xcYuntu);
    }


    /**
     * 数据迁移需要的一些接口
     */

    /**
     * 设置设备总量
     */
    static async _setDeviceCount() {
        //用xcRedis

    }

    /**
     * 添加新设备
     * 数据库是隔离的，需要从scm导入设备
     * 分销管理软件入库  -- scm imeiListByAgentId -- importDevices批量入库
     */
    static async importDevices(imeiList) {
        _setDeviceCount();
    }

    /**
     * 获取当前设备对应的代理商id
     */
    async agentIdByImei() {
        return 2;
    }

    /**
     * 从数据库更新imei-agentId对应关系
     */
    async setImeiAgents() {

    }

    /**
     * 配置设备服务的基本信息
     */
    static async initService(json) {
        //todo

    }

    /**
     * 从json初始化
     * @param {*} json 
     */
    constructor(json) {

        this.imei = json.imei;

        //信号值
        this.gsmSignal = json.gsmSignal;
        //设防状态
        this.defend = json.defend;
        //电门状态
        this.acc = json.acc;
        //经纬度
        this.lat = json.lat;
        this.lng = json.lng;
        //电池仓锁
        this.batteryLock = json.batteryLock;
        // 电池链接
        this.batteryConnect = json.batteryConnect;
        //车锁状态
        this.backWheelLock = json.backWheelLock;
        //是否开启自动设防
        this.isAutoLock = json.isAutoLock;
        //自动设防周期，分钟
        this.autoLockPeriod = json.autoLockPeriod;
        //设备对应的代理商id
        this.agentId = json.agentId;
        //设备状态
        this.state = json.state;
        //设备种类
        this.objectType = json.objectType;
    }

    /**
     * 从redis查询设备状态
     * 如果没有查到设备，则返回null
     * 保证此函数获取的所有设备数据是实时可信的
     */
    static async instanceFromRedis(imei) {
        let ret = await xcRedis.hgetall(`device_info_${imei}`);
        if (ret && ret.imei) {
            let device = new XCEBikeDevice(ret);
            //获取设备状态
            await device.getState();
            //todo

            return device;
        } else {
            return;
        }

    }

    /**
     * 将信息组装成json
     */
    jsonData() {
        let json = {};

        json.imei = this.imei;
        //信号值
        json.gsmSignal = this.gsmSignal;
        //设防状态
        json.defend = this.defend;
        //电门状态
        json.acc = this.acc;
        //经纬度
        json.lat = this.lat;
        json.lng = this.lng;
        //电池仓锁
        json.batteryLock = this.batteryLock;
        // 电池链接
        json.batteryConnect = this.batteryConnect;
        //车锁状态
        json.backWheelLock = this.backWheelLock;
        //是否开启自动设防
        json.isAutoLock = this.isAutoLock;
        //自动设防周期，分钟
        json.autoLockPeriod = this.autoLockPeriod;
        //设备对应的代理商id
        json.agentId = this.agentId;
        //设备状态
        //json.state = this.state;
        //设备种类
        json.objectType = this.objectType;
        return json;
    }

    /**
     * 将信息组装成mapData
     */
    mapData() {
        let map = new Map();
        map.set('imei', this.imei);
        //信号强度
        map.set('gsmSignal', this.gsmSignal);
        //设防状态
        map.set('defend', this.defend);
        //电门状态
        map.set('acc', this.acc);
        //经纬度
        map.set('lat', this.lat);
        map.set('lng', this.lng);
        //电池仓锁
        map.set('batteryLock', this.batteryLock);
        // 电池链接
        map.set('batteryConnect', this.batteryConnect);
        ////车锁状态
        map.set('backWheelLock', this.backWheelLock);
        //是否开启自动设防
        map.set('isAutoLock', this.isAutoLock);
        //自动设防周期，分钟
        map.set('autoLockPeriod', this.autoLockPeriod);
        //设备对应的代理商id
        map.set('agentId', this.agentId);
        //设备种类
        map.set('objectType', this.objectType);

        return map;
    }

    /**
     * 保存设备信息到Redis
     * boolean, 数字存储时会被转为字符串
     * 只会对其中非空的字段进行赋值
     * 空字段不会影响Redis已有的对应字段数据
     */
    async saveDeviceInfo() {
        let ret = await xcRedis.hmset(`device_info_${this.imei}`, this.mapData());
        return ret;
    }

    /**
     * 从后台更新车辆的可预约时长
     */
    static async updateBikeBookingTime() {

    }

    /**
     * 设备绑定一辆车
     * 同时往车-imei表中记录一条记录
     * carId
     * @param {*} carId 
     */
    async bindCarId(carId) {
        let ret = await xcRedis.set(`imeiCarBindings_${this.imei}`, carId);
        ret = await xcRedis.set(`carImeiBindings_${carId}`, this.imei);
        return ret;
    }

    /**
     * 解绑车-imei
     * 同时解绑imei-车
     * @param {*} carId 
     */
    async unBindCarId(carId) {
        let ret = await xcRedis.del(`imeiCarBindings_${this.imei}`);
        ret = await xcRedis.del(`carImeiBindings_${carId}`);
        return ret;
    }

    /**
     * 返回设备绑定的车Id
     */
    async bindedCarId() {
        let ret = await xcRedis.get(`imeiCarBindings_${this.imei}`);
        return ret;
    }

    /**
     * 给定carId查找绑定的设备imei
     * @param {*} carId 
     */
    static async imeiByCarId(carId) {
        let ret = await xcRedis.get(`carImeiBindings_${carId}`);
        return ret;
    }

    /**
     * 获取当前设备状态
     */
    async getState() {

        if (await this.bookingUsrId()) { //有预约信息
            this.state = XC_EBIKE_DEVICE_STATE.BOOKING;
            return XC_EBIKE_DEVICE_STATE.BOOKING;
        } else {
            let ret = await xcRedis.get(`device_state_${this.imei}`);
            this.state = ret;
            return ret;
        }

    }

    async isRiding() {
        const state = await this.getState();
        return state === XC_EBIKE_DEVICE_STATE.RIDING;
    }

    async isChangingBattery() {
        const state = await this.getState();
        return state === XC_EBIKE_DEVICE_STATE.CHANGING_BATTERY;
    }

    /**
     * 设置设备状态
     */
    async setState(state) {
        this.state = state;
        let ret = await xcRedis.set(`device_state_${this.imei}`, state);
        return ret;
    }

    /**
     * 获取预约了该设备的用户id
     * 有过期时间:EBikeBookingTime
     */
    async bookingUsrId() {
        let ret = await xcRedis.get(`deviceBooking_${this.imei}`);
        return ret;
    }

    /**
     * 预约车辆-设置用户预约key：usrId，有过期时间:EBikeBookingTime
     * @param {} usrId 
     */
    async setBookingUsrId(usrId) {
        let ret = await xcRedis.set(`deviceBooking_${this.imei}`, usrId, 'EX', EBikeBookingTime);
        return ret;
    }

    /**
     * 取消预约-删除预约设备的用户id
     * @param {*} usrId 
     */
    async delBookingUsrId(usrId) {
        let ret = await xcRedis.del(`deviceBooking_${this.imei}`);
        return ret;
    }

    /**
     * 处理预约车过期事件
     * @param {*} msg 
     * @param {*} key 
     */
    static handleDeviceBookingExpired(msg, key){
        let reg = new RegExp(`${xcRedis.options.keyPrefix}deviceBooking_(\\d{15})`);
        if(reg.test(key)){//拿到设备预约过期key
            //提取imei做过期处理
            let imei = RegExp.$1;
            let device = new XCEBikeDevice({imei:imei});
            device.setState(XC_EBIKE_DEVICE_STATE.READY);//置设备状态
        }

    }

    /**
     * 设置车辆的骑行人
     */
    async setRidingUsrId(usrId) {
        let ret = await xcRedis.set(`deviceRidingUsrId_${this.imei}`, usrId);
        return ret;
    }

    /**
     * 获取设备骑行人
     */
    async ridingUsrId() {
        let ret = await xcRedis.get(`deviceRidingUsrId_${this.imei}`);
        return ret;
    }

    /**
     * 删除设备骑行人
     */
    async delRidingUsrId() {
        let ret = await xcRedis.del(`deviceRidingUsrId_${this.imei}`);
        return ret;
    }


    /**
     * 获取某个代理商的设备总量
     */
    static async deviceCount(agentId) {
        let ret = await xcRedis.get(`${agentId}_deviceCount`);
        return ret;
    }

    /**
     * 获取异常设备总量
     */

    static async deviceAlarmCount(agentId) {
        let ret = await xcRedis.scard(`${agentId}_alarmDevices`);
        return ret;
    }

    /**
     * 设备报警时调用此接口，将设备写入redis：${agentId}_alarmDevices用于计数
     * 工单处理完成接口中将imei移除
     */
    async deviceAlarm() {
        let agentId = await this.agentIdByImei();
        let ret = await xcRedis.sadd(`${agentId}_alarmDevices`, this.imei);
        return ret;
    }

    /**
     * 工单处理完成，解除设备的alarm状态
     */
    async deviceAlarmResolved() {
        let agentId = await this.agentIdByImei();
        let ret = await xcRedis.srem(`${agentId}_alarmDevices`, this.imei);
        return ret;
    }

    /**
     * 获取报修设备总量
     */
    static async deviceBrokenCount(agentId) {
        let ret = await xcRedis.scard(`${agentId}_2FixDevices`);
        return ret;
    }

    /**
     * 设备报修
     */
    async deviceReport2Fix() {
        let agentId = await this.agentIdByImei();
        let ret = await xcRedis.sadd(`${agentId}_2FixDevices`, this.imei);
        return ret;
    }

    /**
     * 设备检修OK
     */
    async deviceFixed() {
        let agentId = await this.agentIdByImei();
        let ret = await xcRedis.srem(`${agentId}_2FixDevices`, this.imei);
        return ret;
    }

    /**
     * 离线设备总数
     */
    static async deviceOfflineCount(agentId) {
        let ret = await xcRedis.scard(`${agentId}_offlineDevices`);
        return ret;
    }


    /**
     * 设备离线
     */

    async deviceOffline() {
        //保存当前状态
        await this._snapshotState();
        //更新状态
        await this.setState(XC_EBIKE_DEVICE_STATE.OFFLINE);
        //设备离线
        let agentId = await this.agentIdByImei();
        let ret = await xcRedis.sadd(`${agentId}_offlineDevices`, this.imei);
        return ret;
    }

    /**
     * 设备上线
     */

    async deviceOnline() {
        //设备上线后恢复离线前状态
        await this._retriveState();
        let agentId = await this.agentIdByImei();
        let ret = await xcRedis.srem(`${agentId}_offlineDevices`, this.imei);
        return ret;
    }

    /**
     * 保存设备离线前的状态
     */
    async _snapshotState() {
        //如果设备处于预约状态下掉线，则取消预约
        if (this.state == XC_EBIKE_DEVICE_STATE.BOOKING) {
            this.state = XC_EBIKE_DEVICE_STATE.READY;
        }
        let ret = await xcRedis.set('deviceStateBeforeOffline', this.state);
        return ret;
    }

    /**
     * 恢复设备离线前状态
     */
    async _retriveState() {
        let ret = await xcRedis.get('deviceStateBeforeOffline');
        await this.setState(this.state);
    }


    /**
     * 正常设备总数
     * 计算得出
     */
    static async deviceRegularCount(agentId) {
        let deviceCount = await XCEBikeDevice.deviceCount(agentId);
        let brokenCount = await XCEBikeDevice.deviceBrokenCount(agentId);
        let alarmCount = await XCEBikeDevice.deviceAlarmCount(agentId);
        let offlineCount = await XCEBikeDevice.deviceOfflineCount(agentId);

        let regularCount = deviceCount - brokenCount - alarmCount - offlineCount;
        return regularCount;
    }

    /**
     * 新定位点上报处理
     */
    async rcvdGPS() {
        let ret = await this.updateItinerary();
        return ret;
    }

    /**
     * 最小移动距离
     */
    static get DIST_MIN() {
        return 15;
    }

    /**
     * 根据GPS点更新自己在redis中的行程信息、计费信息
     * 依赖以下接口 itineraryInfo、itineraryMapData、saveItineraryInfo2Redis
     */
    async updateItinerary(lat, lng, timestamp) {
        await this.getItineraryInfo();
        if (this.itineraryInfo.startTime === 0) {
            this.itineraryInfo.startTime = timestamp;
            this.itineraryInfo.startLat = lat;
            this.itineraryInfo.startLng = lng;
            this.itineraryInfo.curTime = timestamp;
            this.itineraryInfo.curLat = lat;
            this.itineraryInfo.curLng = lng;
            this.itineraryInfo.curItinerary = 0;
            this.itineraryInfo.curCost = await this.computeCost();
        } else {
            const dist = this.getDistanceGPS(lat, lng, this.itineraryInfo.curLat, this.itineraryInfo.curLng);
            this.itineraryInfo.curTime = timestamp;
            this.itineraryInfo.curLat = lat;
            this.itineraryInfo.curLng = lng;
            if (dist > XCEBikeDevice.DIST_MIN) {
                this.itineraryInfo.curItinerary += dist;
            }
            this.itineraryInfo.curCost = await this.computeCost();
        }
        await this.saveItineraryInfo2Redis();
    }

    getDistanceGPS(lat1, lng1, lat2, lng2) {
        var radLat1 = toRadians(lat1);
        var radLat2 = toRadians(lat2);
        var deltaLat = radLat1 - radLat2;
        var deltaLng = toRadians(lng1) - toRadians(lng2);

        var dis = 2 * Math.asin(Math.sqrt(Math.pow(Math.sin(deltaLat / 2), 2) + Math.cos(radLat1) * Math.cos(radLat2) * Math.pow(Math.sin(deltaLng / 2), 2)));

        return dis * 6378137;

        function toRadians(d) {
            return d * Math.PI / 180;
        }
    }

    async computeCost() {
        const startDist = 5; // 5分钟后开始收费
        const startTime = 2; // 2 min
        const fixedDist = 5e3; // 5 km
        const fixedDistCost = 2e3; // 2 RMB
        const fixedTime = 1 * 60 * 60; // 1 hour
        const fixedTimeCost = 0; // 一小时内不收钱
        const costPerMeter = 1; // 1 RMB per mile (1 cent per meter)
        const costPerHour = 5e2; // 0.5 RMB per hour
        const penaltyDist = 1e3; // 1 km
        const smallPenalty = 10e3; // 10 RMB
        const largePenalty = 100e3; // 100 RMB

        const itinerary = this.itineraryInfo.curItinerary;
        const duration = moment.unix(this.itineraryInfo.curTime).diff(moment.unix(this.itineraryInfo.startTime), "seconds");
        let distCost = 0;
        let timeCost = 0;

        distCost = fixedDistCost + Math.ceil(itinerary - fixedDist) * costPerMeter;
        timeCost = fixedTimeCost + Math.ceil((duration - fixedTime) / (60 * 60)) * costPerHour;

        // use fixed cost if fixed limit is not exceeded.
        if (this.itineraryInfo.curItinerary < fixedDist) {
            distCost = fixedDistCost;
        }
        if (duration < fixedTime) {
            timeCost = fixedTimeCost;
        }

        // clear the cost if start limit is not exceeded.
        if (this.itineraryInfo.curItinerary < startDist) {
            distCost = 0;
        }
        if (duration < startTime) {
            timeCost = 0;
        }

        return distCost + timeCost;
    }

    /**
     * 从Redis中获取已有行程信息
     */
    async getItineraryInfo() {
        let ret = await xcRedis.hgetall(`device_itinerary_info_${this.imei}`);

        function parseStringInt(s) {
            const i = parseInt(s);
            if (isNaN(i)) {
                return 0;
            } else {
                return i;
            }
        }

        function parseStringFloat(s) {
            const f = parseFloat(s);
            if (isNaN(f)) {
                return 0.0;
            } else {
                return f;
            }
        }

        ret.startTime = parseStringInt(ret.startTime);
        ret.startLat = parseStringFloat(ret.startLat);
        ret.startLng = parseStringFloat(ret.startLng);
        ret.curTime = parseStringInt(ret.curTime);
        ret.curLat = parseStringFloat(ret.curLat);
        ret.curLng = parseStringFloat(ret.curLng);
        ret.curItinerary = parseStringFloat(ret.curItinerary);
        ret.curCost = parseStringFloat(ret.curCost);
        ret.endTime = parseStringInt(ret.endTime);

        this.itineraryInfo = ret;
        return ret;
    }

    /**
     * 行程信息json2map
     */
    itineraryMapData() {
        let map = new Map();
        map.set('startTime', this.itineraryInfo.startTime);
        map.set('startLat', this.itineraryInfo.startLat);
        map.set('startLng', this.itineraryInfo.startLng);

        map.set('endTime', this.itineraryInfo.curTime);
        map.set('curLat', this.itineraryInfo.curLat);
        map.set('curLng', this.itineraryInfo.curLng);
        map.set('curItinerary', this.itineraryInfo.curItinerary);
        map.set('curCost', this.itineraryInfo.curCost);
        //other data to do

        return map;
    }

    /**
     * 保存计算好的行程信息
     */
    async saveItineraryInfo2Redis() {
        let ret = await xcRedis.hmset(`device_itinerary_info_${this.imei}`, this.itineraryMapData());
        return ret;
    }

    /**
     * 获取设备行程表DAO
     */
    async deviceItineraryDAO() {
        let model = xcSequelize.define(`xc_ebike_device_itinerary_${this.imei}`, XC_EBIKE_DEVICE_ITINERARY_TABLE_DS, CREATE_RULE);
        await model.sync();
        return model;
    }

    /**
     * 将生成好的行程信息入库
     */
    async produceItinerary() {
        await this.getItineraryInfo();
        let deviceItinerary = await this.deviceItineraryDAO();
        let deviceItineraryInfo = {
            startTime: moment(this.itineraryInfo.startTime * 1000).format('YYYY-MM-DD HH:mm:ss'),
            startLat: this.itineraryInfo.startLat,
            startLng: this.itineraryInfo.startLng,
            endTime: moment(this.itineraryInfo.endTime * 1000).format('YYYY-MM-DD HH:mm:ss'),
            endLat: this.itineraryInfo.curLat,
            endLng: this.itineraryInfo.curLng,
            itinerary: this.itineraryInfo.curItinerary
        };
        let newItineraryInfo = await deviceItinerary.create(deviceItineraryInfo);
        newItineraryInfo.imei = this.imei;
        newItineraryInfo.cost = this.itineraryInfo.curCost;

        return newItineraryInfo;
    }

    /**
     * 根据行程id获取行程信息
     * @param {设备行程记录表里的id} itinId 
     */
    async getItineraryByItinId(itinId) {
        let deviceItinerary = await this.deviceItineraryDAO();
        let findByItinId = await deviceItinerary.findOne({
            where: {
                itinId: {
                    $eq: Number(itinId)
                }
            }
        });
        if (findByItinId === null || findByItinId === undefined) {
            return null;
        }
        if (findByItinId.hasOwnProperty('dataValues')) {
            return findByItinId.dataValues;
        } else {
            return null;
        }
    }

    /**
     * 将设备的行程信息重置
     */
    async resetIntinerary() {
        await this.getItineraryInfo();
        let map = this.itineraryMapData();
        map.set('startTime', 0);
        let ret = await xcRedis.hmset(`device_itinerary_info_${this.imei}`, map);
        return ret;
    }

    /**
     * 将设备的GPS信息更新到yuntu
     */
    static async update2Yuntu(imei, lng, lat) {
        const name = `${imei}`;
        const location = `${lng},${lat}`;

        let tableid = xcYuntu.deviceTableid;
        return await XCYuntuUtil.DataUpdate(xcYuntu, tableid, name, location);
    }

    /**
     * 将设备的GPS信息从yuntu删除
     */
    static async deleteFromYuntu(imei) {
        const name = `${imei}`;
        
        let tableid = xcYuntu.deviceTableid;
        return await XCYuntuUtil.DataDelete(xcYuntu, tableid, name);
    }

    /**
     * 根据所给的GPS在yuntu中查找附近的设备，并返回数据列表
     * datas: [ {   _id: '1',
                    _location: '114.525,30.511',
                    _name: '865067022403441',
                    _address: '湖北省武汉市江夏区丁姑山',
                    _createtime: '2017-12-02 16:12:54',
                    _updatetime: '2017-12-02 16:17:49',
                    _province: '湖北省',
                    _city: '武汉市',
                    _district: '江夏区',
                    _distance: '0',
                    _image: [] } ]
     */
    static async searchFromYuntu(lng, lat, radius) {
        const location = `${lng},${lat}`;

        let tableid = xcYuntu.deviceTableid;
        return await XCYuntuUtil.DataSearchAround(xcYuntu, tableid, location, radius);
    }
}

module.exports = XCEBikeDevice;