/**
 * 电子围栏类
 */
const Sequelize = require('sequelize');
const XCYuntuUtil = require('./XCYuntuUtil');

let xcSequelize = null;
let xcRedis = null;
let xcYuntu = null;

const CREATE_RULE = {
    freezeTableName: true,
    paranoid: true
}; //添加删除deletedAt

//电子围栏类型
let GFENCE_TYPE = {
    AtService: '1',//服务区
    Usable: '2',//租还区
    Denied: '3',//禁行区
};

//电子围栏形状
const SHAPE_TYPE = {
    polygon: 'polygon',
    circle: 'circle'
};

const XC_EBIKE_GFENCE_DS = {
    id: {
        type: Sequelize.INTEGER(8).UNSIGNED,
        primaryKey: true,
        allowNull: false,
        autoIncrement: true
    },
    name: { //添加用户新建时的区域名字
        type: Sequelize.STRING(128),
        unique: true,
        allowNull: false
    },
    type: {//电子围栏类型
        type: Sequelize.STRING(16),
        allowNull: false,
    },
    shapeType: {//电子围栏形状
        type: Sequelize.STRING(16),
        allowNull: false
    },
    centerLat: {
        type: Sequelize.DOUBLE(9, 6),
        allowNull: false
    },
    centerLng: {
        type: Sequelize.DOUBLE(9, 6),
        allowNull: false,
    },
    pointList: { //电子围栏顶点列表。json串
        type: Sequelize.STRING(1024),
        allowNull: false
    }

}

class XCEBikeGFence {

    /**
     * 配置redis client
     */
    static configRedis(redisClient, redisSubClient) {
        xcRedis = redisClient;
    }

    static get SHAPE_TYPE() {
        return SHAPE_TYPE;
    }

    static get GFENCE_TYPE(){
        return GFENCE_TYPE;
    }
    /**
     * 配置数据库实例
     */
    static configSequelize(sequelizeInstance) {
        xcSequelize = sequelizeInstance;
    }

    /**
     * 配置yuntu实例
     */
    static async configYuntu(yuntu) {
        xcYuntu = yuntu;
        await XCYuntuUtil.gFenceTableidEnsureInit(yuntu);
    }

    static log() {
        console.log(xcRedis, xcSequelize, xcYuntu);
    }

    /**
     * 
     * @param {*} json 
     * agentId，必填
     */
    constructor(json) {
        //电子围栏类型
        this.type = json.type;
        //电子围栏所属的agentId
        this.agentId = json.agentId;
        //电子围栏id
        this.gFenceId = json.gFenceId;
        //电子围栏name
        this.name = json.name;
        //电子围栏形状
        this.shapeType = json.shapeType;
        //电子围栏中心点
        this.centerLng = json.centerLng;
        this.centerLat = json.centerLat;
        //电子围栏点列表{lat:lat,lng:lng}
        this.pointList = json.pointList;
        this._DAO = null;
        if(this.agentId) {
            this.gFenceDAO(`xc_ebike_gfence_${this.agentId}`);
        } else {
            throw 'gFence no agentId';
        }
    }

    /**
     * 电子围栏信息转mapData
     */
    mapData() {
        let map = new Map();
        map.set('gFenceId', this.gFenceId);
        map.set('type', this.type);
        map.set('name', this.name);
        map.set('shapeType', this.shapeType);
        map.set('centerLng', this.centerLng);
        map.set('centerLat', this.centerLat);
        map.set('pointList', this.pointList);
        map.set('agentId', this.agentId);

        return map;
    }

    /**
     * 组合地理围栏的Redis Key
     * @param {*} agentId 
     * @param {*} type 
     * @param {*} gFenceId 
     */
    static gFenceKey(agentId, gFenceId) {
        return `gFence_${agentId}_${gFenceId}`;
    }

    /**
     * 从Redis中初始化一个电子围栏
     * 没有查询到电子围栏信息则返回null
     */
    static async instanceFromRedis(agentId, gFenceId) {
        let json = await xcRedis.hgetall(XCEBikeGFence.gFenceKey(agentId, gFenceId));
        if (json && json.gFenceId) {
            let gFence = new XCEBikeGFence(json);
            return gFence;
        } else {
            return;
        }
    }

    /**
     * 将围栏保存在Redis中
     */
    async save2Redis() {
        let mapData = this.mapData();
        let ret = await xcRedis.hmset(XCEBikeGFence.gFenceKey(this.agentId, this.gFenceId), mapData);
        return ret;
    }

    /**
     * 将电子围栏从Redis中删除
     */
    async delFromRedis(){
        let ret = await xcRedis.del(XCEBikeGFence.gFenceKey(this.agentId, this.gFenceId));
        return ret;
    }

  
     /* 电子围栏建立基类
     * @param {*} tableName 
     */
    async gFenceDAO(tableName) {
        this._DAO = xcSequelize.define(tableName, XC_EBIKE_GFENCE_DS, CREATE_RULE);
        return this._DAO;
    }

    /**
     * 创建电子围栏数据
     * @param {*}
     */
    async gFenceCreate() {
        let gFenceDAO = this._DAO;
        await gFenceDAO.sync();
        let usableCreate = await gFenceDAO.upsert({
            name: this.name,
            shapeType: this.shapeType,
            type: this.type,
            centerLat: this.centerLat,
            centerLng: this.centerLng,
            pointList: this.pointList,
            deletedAt: null
        });
        let findOne = await gFenceDAO.findAll({
            where: {
                name: {
                    $eq: this.name
                },
                type: {
                    $eq: this.type
                }
            }
        });
        this.gFenceId = findOne[0].dataValues.id;
        let save2Redis = await this.save2Redis();        
        return findOne;
    }

    /**
     * 获取电子围栏数据
     * @param {*}  
     */
    async gFenceGet() {
        let gFenceDAO = this._DAO;
        await gFenceDAO.sync();
        let gFenceFind = await gFenceDAO.findAll({
            where: {
                deletedAt:{
                    $eq:null
                },
                type: {
                    $eq: this.type
                }
            }
        });
        return gFenceFind;
    }

    /**
     * 更新电子围栏数据
     */
    async gFenceUpdate() {
        let gFenceDAO = this._DAO;
        await gFenceDAO.sync();
        let gFenceUpdate = await gFenceDAO.update({
            name: this.name,
            shapeType: this.shapeType,
            centerLat: this.centerLat,
            centerLng: this.centerLng,
            pointList: this.pointList,
            deletedAt: null
        }, {
                where: {
                    id: {
                        $eq: Number(this.gFenceId)
                    },
                    type: {
                        $eq: this.type
                    }
                }
            }
        );
        let save2Redis = await this.save2Redis();        
        return gFenceUpdate;
    }

    /**
     * 删除id对应的租还区数据
     */
    async gFenceDestroy() {
        let gFenceDAO = this._DAO;
        await gFenceDAO.sync();
        let gFenceDestroy = await gFenceDAO.destroy({
            where: {
                id: {
                    $eq: Number(this.gFenceId)
                },
                type: {
                    $eq: this.type
                }
            }
        });
        let deleteRedis = await this.delFromRedis();
        return gFenceDestroy;
    }

    /**
     * 将电子围栏的中心点GPS信息更新到yuntu
     */
    static async update2Yuntu(gFenceType, gFenceId, lng, lat) {
        const name = `${gFenceId}`;
        const location = `${lng},${lat}`;

        let tableid;
        if(gFenceType == GFENCE_TYPE.AtService) {
            tableid = xcYuntu.gFenceServiceTableid;
        } else if(gFenceType == GFENCE_TYPE.Usable) {
            tableid = xcYuntu.gFenceUsableTableid;
        }
        return await XCYuntuUtil.DataUpdate(xcYuntu, tableid, name, location);
    }

    /**
     * 将电子围栏的中心点GPS信息从yuntu删除
     */
    static async deleteFromYuntu(gFenceType, gFenceId) {
        const name = `${gFenceId}`;
        let tableid;
        if(gFenceType == GFENCE_TYPE.AtService) {
            tableid = xcYuntu.gFenceServiceTableid;
        } else if(gFenceType == GFENCE_TYPE.Usable) {
            tableid = xcYuntu.gFenceUsableTableid;
        }
        return await XCYuntuUtil.DataDelete(xcYuntu, tableid, name);
    }

    /**
     * 根据所给的GPS在yuntu中查找附近的电子围栏，并返回数据列表
     * datas: [ {   _id: '1',
                    _location: '114.525,30.511',
                    _name: '1',
                    _address: '湖北省武汉市江夏区丁姑山',
                    _createtime: '2017-12-02 16:12:54',
                    _updatetime: '2017-12-02 16:17:49',
                    _province: '湖北省',
                    _city: '武汉市',
                    _district: '江夏区',
                    _distance: '0',
                    _image: [] } ]
     */
    static async searchFromYuntu(gFenceType, lng, lat, radius) {
        const location = `${lng},${lat}`;

        let tableid;
        if(gFenceType == GFENCE_TYPE.AtService) {
            tableid = xcYuntu.gFenceServiceTableid;
        } else if(gFenceType == GFENCE_TYPE.Usable) {
            tableid = xcYuntu.gFenceUsableTableid;
        }
        return await XCYuntuUtil.DataSearchAround(xcYuntu, tableid, location, radius);
    }
}


module.exports = XCEBikeGFence;