/**
 * 工单数据
 * 维修表单
 * 异常表单
 */

const Sequelize = require('sequelize');
const CREATE_RULE = {
    freezeTableName: true
}

/**
 * 维修类型
 */
const FIX_TYPE = {
    LOCK: 0,
    BATTERY: 1,
    //to do
    //,
    OTHER: 3,
}

/**
 * 维修状态
 */
const FIX_STATE = {
    TO_FIX: 0,
    FIXING: 1,
    FIXED: 2,
    //返厂维修为3
}


const XCEBikeFixTicketStruct = {
    ticketNo: {
        type: Sequelize.INTEGER(32).UNSIGNED,
        primaryKey: true,
        allowNull: false,
        autoIncrement: true
    },
    imei: {
        type: Sequelize.STRING(16),
        allowNull: false
    },
    ebikeNo: {
        type: Sequelize.STRING(32),
        allowNull: false,
    },
    type: {//维修类型
        type: Sequelize.INTEGER.UNSIGNED,
        allowNull: false
    },
    fixReason: {
        type: Sequelize.STRING(1024)
    },
    photo: {
        type: Sequelize.STRING(128)
    },
    reportUsrId: {
        type: Sequelize.STRING(64),
        allowNull: false
    },
    state: {//工单状态
        type: Sequelize.INTEGER.UNSIGNED,
        allowNull: false,
        defaultValue: FIX_STATE.TO_FIX,
    },
    fixManId: {
        type: Sequelize.STRING(64),
    },
    fixedTime: {
        type: Sequelize.DATE,
    },
    extraInfo: {
        type: Sequelize.STRING(1024),
    }

};


const ALARM_TYPE = {
    LOW_BATTERY: 0,
    HIGHT_BATTERY: 1,
    MOVE_ALARM: 2,
    OUT_GFENCE: 3,
    TIME_2_CHECK: 4,
    POWER_CUT: 5,
}

/**
 * 异常工单
 */

const XCEBikeAlarmTicketStruct = {
    ticketNo: {
        type: Sequelize.INTEGER(32).UNSIGNED,
        primaryKey: true,
        allowNull: false,
        autoIncrement: true
    },
    imei: {
        type: Sequelize.STRING(16),
        allowNull: false
    },
    ebikeNo: {
        type: Sequelize.STRING(32),
        allowNull: false,
    },
    type: {//异常类型
        type: Sequelize.INTEGER.UNSIGNED,
        allowNull: false
    },
    state: {//工单状态       
        type: Sequelize.INTEGER.UNSIGNED,
        allowNull: false,
        defaultValue: FIX_STATE.TO_FIX,
    },
    fixManId: {
        type: Sequelize.STRING(64),
    },
    fixedTime: {
        type: Sequelize.DATE,
    },
    extraInfo: {
        type: Sequelize.STRING(1024),
    }
};


let xcSequelize = null;
let xcRedis = null;

class XCEBikeTicket {

    static get FIX_TYPE(){
        return FIX_TYPE;
    }

    static get FIX_STATE(){
        return FIX_STATE;
    }

    static get ALARM_TYPE(){
        return ALARM_TYPE;
    }


    /**
    * 配置redis client
    */
    static configRedis(redisClient) {
        xcRedis = redisClient;
    }

    /**
     * 配置数据库实例
     */
    static configSequelize(sequelizeInstance) {
        xcSequelize = sequelizeInstance;
    }

    static log() {
        console.log(xcRedis, xcSequelize);
    }

    constructor(agentId, tableName, tableStruct) {
        this.agentId = agentId;
        this._DAO = xcSequelize.define(tableName, tableStruct, CREATE_RULE);
    }

    /**
     * 创建一条工单/FixTicket or alarmTicket
     * @param {Object}       json
     * json(fixJson/alarmJson) include essential param:
     * fixJson:
     * @param {string 16}    imei
     * @param {string 32}    ebikeNo
     * @param {unsigned int} type
     * @param {unsigned int} state
     * default FIX_STATE.TO_FIX
     * @param {string 64}    reportUsrId
     * 
     * alarmJson:
     * @param {string 16}    imei
     * @param {string 32}    ebikeNo
     * @param {unsigned int} type
     * ALARM_TYPE
     * @param {unsigned int} state
     * default FIX_STATE.TO_FIX
     */
    async create(json) {
        await this._DAO.sync();
        await this._DAO.create(json);
    }

    /**
     * 标记一条工单结束
     * @param {*} param0 
     */
    async handled(ticketNo) {
        await this._DAO.update({
            fixedTime: new Date(),
            state: FIX_STATE.FIXED
        }, {
                where: {
                    ticketNo: {
                        $eq: Number(ticketNo)
                    }
                }
            }
        );
        return;
    }
    /**
     * 标记一条工单正在被处理
     * @param {*} ticketNo 
     * @param {*} fixManId 
     */
    async updateInfo(ticketNo,fixManId) {
        let updateTicketInfo = await this._DAO.update({
            fixManId: fixManId,
            state: FIX_STATE.FIXING
        },{
            where :{
                ticketNo: {
                    $eq : Number(ticketNo)
                }
            }
        });
        return updateTicketInfo;
    }

    /**
     * 获取一条工单信息
     * @param {*} ticketNo 
     */
    async ticketInfoInDB(ticketNo) {
        let ticketInfo = await this._DAO.findOne({
            where: {
                ticketNo: {
                    $eq: Number(ticketNo)
                }
            }
        });
        return ticketInfo;
    }

    /** 
     * 工单列表
      分页查询
      按状态查询
      按维修人查询
     */

    /**
     * @param {*} countPerpage  每页数量 >= 1
     * @param {*} currPage 当前页数 >= 1
     */
    async ticketListState(countPerpage, currPage, state) {
        let pageInfo = await this._DAO.findAll({
            where: {
                state: state
            },
            'limit': countPerpage,
            'offset': countPerpage * (currPage - 1)
        });
        return pageInfo;
    }

    async ticketListFixMan(countPerpage, currPage, fixManId) {
        let pageInfo = await this._DAO.findAll({
            where: {
                fixManId: fixManId
            },
            'limit': countPerpage,
            'offset': countPerpage * (currPage - 1)
        });
        return pageInfo;
    }

    async ticketList(countPerpage, currPage) {
        let pageInfo = await this._DAO.findAll({
            'limit': countPerpage,
            'offset': countPerpage * (currPage - 1)
        });
        return pageInfo;
    }

}
/**
 * 维修工单子类
 */

class XCEBikeFixTicket extends XCEBikeTicket {
    constructor(agentId) {
        if (agentId) {
            let XC_EBIKE_FIX_TICKET_TABLE = `xc_ebike_fix_tickets_${agentId}`;
            super(agentId, XC_EBIKE_FIX_TICKET_TABLE, XCEBikeFixTicketStruct)
        } else {
            throw 'XCEBikeFixTicket.daoWithAgentId agentId null';
        }
    }
}
/**
 *写异常子类
 */

class XCEBikeAlarmTicket extends XCEBikeTicket {
    constructor(agentId) {
        if (agentId) {
            let XC_EBIKE_ALARM_TICKET_TABLE = `xc_ebike_alarm_tickets_${agentId}`;
            super(agentId, XC_EBIKE_ALARM_TICKET_TABLE, XCEBikeAlarmTicketStruct);
        } else {
            throw 'XCEBikeAlarmTicket.daoWithAgentId agentId null';
        }
    }
}

module.exports = {
    XCEBikeTicket,
    XCEBikeFixTicket,
    XCEBikeAlarmTicket
}
