/**
 * XCEBikeUser.js
 * 用户信息及操作类
 * 只有操作方法，无操作策略。操作策略在service层维护
 * lyoncat
 */

/**
 * 用户状态表
 */
const XC_EBIKE_USER_STATE = {
    SIGN_UP: '1',
    AUTHED: '2',
    READY: '3',
    BOOKING: '4',
    LEAVING: '5',
    RIDING: '6',
    TO_PAY: '7',
}

const CREATE_RULE = {
    freezeTableName: true
}


const Sequelize = require('sequelize');


/**
 * 用户信息表结构定义
 */
const XCEBikeUser_DS = {
    id: {
        type: Sequelize.STRING(64),
        primaryKey: true,
        allowNull: false
    },
    phone: {
        type: Sequelize.STRING(16),
    },
    authed: {
        type: Sequelize.BOOLEAN,
        defaultValue: false
    },
    personInfo: {
        type: Sequelize.STRING(1024),
    },
    deposited: {
        type: Sequelize.BOOLEAN,
        allowNull: false,
        defaultValue: false,
    },
    depositedInfo: {
        type: Sequelize.STRING(1024),
    }

};


/**
 * 可预约时间
 */
const EBikeBookingTime = 5 * 60; //单位秒
// const EBikeBookingTime = 5; //单位秒

let xcSequelize = null;
let xcRedis = null;

/**
 * XCEBikeUser类定义
 */
class XCEBikeUser {
    static get XC_EBIKE_USER_STATE(){
        return XC_EBIKE_USER_STATE;
    }

    static log() {
        console.log(xcRedis, xcSequelize);
    }

    /**
     * 配置redis client
     * 和RedisSubClient，用于监听Key过期事件
     */
    static configRedis(redisClient, redisSubClient) {
        xcRedis = redisClient;
        XCEBikeUser.configExpiredKeyListen(redisSubClient);
    }

    /**
     * 配置预约过期key事件监听
     */
    static configExpiredKeyListen(redisSubClient){
        if(xcRedis && redisSubClient){
            redisSubClient.subscribe(`__keyevent@${xcRedis.options.db}__:expired`, 1);//同一db
            redisSubClient.on('message', (msg, key) => {
                XCEBikeUser.handleExpiredKeyEvents(msg, key);
              })
        }
        
    }

    /**
     * 处理过期消息和过期Key
     */
    static handleExpiredKeyEvents(msg, key){
        XCEBikeUser.handleEbikeBookingExpired(msg, key);
    }

    /**
     * 配置数据库实例
     */
    static configSequelize(sequelizeInstance) {
        xcSequelize = sequelizeInstance;
    }

    /**
     * 使用json数据新创建用户
     * 创建成功或数据表里有该用户，均返回true
     * 抛出异常则创建失败
     */
    static async insertUsr2DB(usrId, agentId, tel) {
        if (usrId && agentId) {
            let dao = XCEBikeUser.getAgentUserTableDAO(agentId);
            await dao.sync();
            let ret = await dao.findOrCreate({
                where: {
                    id: usrId
                },
                defaults: {
                    id: usrId,
                    phone: tel
                }
            });

            //维护redis中全量用户数量
            await xcRedis.sadd(`${agentId}_userCount`, this.id);
            //维护Redis中用户和代理商的关系
            ret = await xcRedis.set(`usrAgentBindings_${this.id}`, agentId);

            return true;

        } else {
            return false;
        }
    }


    constructor(id) {

        this.id = id;
        /**
         * 以下字段用来明确数据结构，不能准确表示实际信息
         * 相应字段的读写操作必须调用对应方法
         */
        //电话
        let json = {};

        this.phone = json.phone;
        //是否实名制
        this.authed = json.authed;
        //个人信息字段
        this.personInfo = json.personInfo;
        //用户状态
        this.state = json.state;
        //是否已交押金
        this.deposited = json.deposited;
        //余额
        this.balance = json.balance;
        // 押金订单信息
        this.depositedInfo = json.depositedInfo;
    }

    /**
     * 给定代理商id获取用户信息DAO
     */
    static getAgentUserTableDAO(agentId) {
        return xcSequelize.define(`xc_ebike_usrs_${agentId}`, XCEBikeUser_DS, CREATE_RULE);
    }

    /**
     * 获取用户信息表DAO对象
     */
    async getUserDAO() {
        let agentId = await this.agentId();
        return xcSequelize.define(`xc_ebike_usrs_${agentId}`, XCEBikeUser_DS, CREATE_RULE);
    }

    /**
     * 获取代理商id
     * @param {*} usrId 
     */
    static async agentIdByUsrId(usrId) {
        let ret = await xcRedis.get(`usrAgentBindings_${usrId}`);
        return ret;
    }

    /**
     * 用户绑定代理商id
     * @param {*} agentId 
     */
    async bindAgent(agentId) {
        let ret = await xcRedis.set(`usrAgentBindings_${this.id}`, agentId);
        return ret;
    }

    /**
     * 获取自己对应的agentId
     */
    async agentId() {
        // let ret = await xcRedis.get(`usrAgentBindings_${this.id}`);
        // return ret;
        return 2;
    }

    /**
     * 从db分页获取用户表
     * @param {*} pageNo 
     * @param {*} pageSize 
     */
    static async userList(agentId, pageNo, pageSize) {

    }

    /**
     * 从redis获取总用户量
     */
    static async userCount(agentId) {
        let userCount = await xcRedis.scard(`${agentId}_userCount`);
        return userCount;
    }

    /**
     * 从redis获取已缴押金用户量
     */
    static async depositUserCount(agentId) {
        let dUsrCnt = await xcRedis.scard(`${agentId}_userDepositCount`);
        return dUsrCnt;
    }

    /**
     * 正在骑行的用户量
     * @param {*} agentId 
     */
    static async ridingUserCount(agentId){
        let ridingUsrCnt = await xcRedis.scard(`${agentId}_userRidingCount`);
        return ridingUsrCnt;
    }

    /**
     * 已预约用户量
     * @param {*} agentId 
     */
    static async bookingUserCount(agentId){
        let ret = await xcRedis.scard(`${agentId}_userBookingCount`);
        return ret;
    }

    /**
     * 实名认证用户量
     */
    static async authedUserCount(agentId){
        let ret = await xcRedis.scard(`${agentId}_userAuthedCount`);
        return ret;
    }

    /**
     * 设置自己骑的设备imei
     */
    async setRidingDevice(imei) {
        let agentId = await this.agentId();
        let ret = await xcRedis.set(`${agentId}_usrRidingDevice_${this.id}`, imei);
        //维护骑行用户数量
        await xcRedis.sadd(`${agentId}_userRidingCount`, this.id);

        return ret;
    }

    /**
     * 返回用户正在骑行的车的imei
     */
    async ridingDevice() {
        let agentId = await this.agentId();
        let ret = await xcRedis.get(`${agentId}_usrRidingDevice_${this.id}`);
        return ret;
    }

    /**
     * 返回用户正在骑行的设备的imei
     */
    async ridingDeviceIMEI() {
        let agentId = await this.agentId();
        let ret = await xcRedis.get(`${agentId}_usrRidingDevice_${this.id}`);
        return ret;
    }

    /**
     * 取消自己与车的骑行关系
     */
    async delRidingDevice() {
        let agentId = await this.agentId();
        let ret = await xcRedis.del(`${agentId}_usrRidingDevice_${this.id}`);
        //维护骑行用户数量
        await xcRedis.srem(`${agentId}_userRidingCount`, this.id);
        return ret;
    }

    /**
     * 从后台更新车辆的可预约时长
     */
    static async updateBikeBookingTime() {

    }


    /**
     * 设置用户的预约信息
     * imei为用户预约的车
     */
    async setBookingImei(imei) {
        let agentId = await this.agentId();
        //有过期时间
        let ret = await xcRedis.set(`${agentId}_userBooking_${this.id}`, imei, 'EX', EBikeBookingTime);
        //维护预约用户数量
        await xcRedis.sadd(`${agentId}_userBookingCount`, this.id);
        return ret;
    }

    /**
     * 用户的预约信息
     * 返回预约的设备Imei
     */
    async getBookingImei() {
        let agentId = await this.agentId();
        let ret = await xcRedis.get(`${agentId}_userBooking_${this.id}`);
        return ret;
    }

    /**
     * 删除用户预约信息
     */
    async delBookingInfo() {
        let agentId = await this.agentId();
        let ret = await xcRedis.del(`${agentId}_userBooking_${this.id}`);
        //维护预约用户数量
        await xcRedis.srem(`${agentId}_userBookingCount`, this.id);
        return ret;
    } 


    /**
     * 获取用户的剩余预约时间
     */
    async getBookingTime(){
        let agentId = await this.agentId();
        let ret = await xcRedis.pttl(`${agentId}_userBooking_${this.id}`);
        return ret;
    }

    /**
     * 处理用户预约车辆过期事件
     * @param {*} msg 
     * @param {*} key 
     */
    static async handleEbikeBookingExpired(msg, key){
        let reg = new RegExp(`${xcRedis.options.keyPrefix}(\\d+)_userBooking_(\\w+)`);
        if(reg.test(key)){//拿到用户预约过期key
            //提取userId做过期处理
            let agentId = RegExp.$1;
            let userId = RegExp.$2;
            let user = new XCEBikeUser(userId);
            //更改用户预约状态
            await user.setState(XC_EBIKE_USER_STATE.READY);
            //修改预约信息及预约数量
            let ret = await user.delBookingInfo();
            return ret;
        }
    }

    /**
     * 用户状态
     */
    async getState() {
        let agentId = await this.agentId();
        let bookingImei = await this.getBookingImei(agentId);
        if (bookingImei) {
            this.state = XC_EBIKE_USER_STATE.BOOKING;
            return XC_EBIKE_USER_STATE.BOOKING;
        } else {
            let ret = await xcRedis.get(`${agentId}_userState_${this.id}`);
            this.state = ret;
            return ret;
        }
    }

    /**
     * 设置用户状态
     */
    async setState(state) {
        let agentId = await this.agentId();
        this.state = state;
        let ret = await xcRedis.set(`${agentId}_userState_${this.id}`, state);
        return ret;
    }

    /**
     * 获取用户完整信息
     * 低速操作
     */
    async loadAllInfoSlow() {
        await this._readDBInfo();
        await this._readRedisInfo();
    }

    /**
     * 读取redis中的用户信息
     * state
     * 
     */
    async _readRedisInfo() {
        this.state = await this.getState();
    }

    /**
     * 获取用户基本信息
     * 低速操作
     */
    async _readDBInfo() {
        let model = await this.getUserDAO();
        let ret = await model.findOne({
            where: {
                id: this.id
            }
        });
        if(ret){
            this.phone = ret.phone;
            this.authed = ret.authed;
            this.personInfo = ret.personInfo;
            this.deposited = ret.deposited;
            this.depositedInfo = ret.depositedInfo;
        }else{
            return;
        }
    }


    /**
     * 获取定金信息
     */
    async getDepositedInfo(depositedInfo) {
        await this._readDBInfo();
        return this.depositedInfo;
    }

    // /**
    //  * 用户已经充值，存定金信息
    //  */
    // async setDepositedInfo(sucOrNot, depositedInfo) {
    //     let userDAO = await this.getUserDAO();
    //     if (sucOrNot) {
    //         await userDAO.update({
    //             deposited: true,
    //             depositedInfo: depositedInfo
    //         }, {
    //             where: {
    //                 id: this.id
    //             }
    //         });
    //     } else {
    //         await userDAO.update({
    //             depositedInfo: depositedInfo
    //         }, {
    //             where: {
    //                 id: this.id
    //             }
    //         });
    //     }

    //     //更新已缴定金用户数量
    //     let agentId = await this.agentId();
    //     await xcRedis.sadd(`${agentId}_userDepositCount`, this.id);
    // }

    
    /**
     * 用户发起取回自己的定金
     */
    async withdrawDepositedMoney(){


        //at the end
        //更新已缴定金用户数量
        let agentId = await this.agentId();
        await xcRedis.srem(`${agentId}_userDepositCount`, this.id);
    }

    /**
     * 用于已被实名验证通过，更新数据
     * 
     * @param {*} userInfo 
     */
    async updateAuthedInfo(userInfo){
        //dao操作

        //更新已实名用户数量
        let agentId = await this.agentId();
        let ret = await xcRedis.sadd(`${agentId}_userAuthedCount`, this.id);
        return ret;
    }

    /**
     * 生成ping++订单后存订单信息
     * @param {*} depositedInfo 
     */
    async setDepositCharge(depositedInfo) {
        let userDAO = await this.getUserDAO();
        await userDAO.update({
            depositedInfo: depositedInfo
        }, {
            where: {
                id: this.id
            }
        });

        return true;
    }

    /**
     * 完成定金支付
     * @param {*} depositedInfo 
     */
    async completeDeposit(depositedInfo) {
        let userDAO = await this.getUserDAO();
        await userDAO.update({
            deposited: true,
            depositedInfo: depositedInfo
        }, {
            where: {
                id: this.id
            }
        });

        //更新已缴定金用户数量
        let agentId = await this.agentId();
        await xcRedis.sadd(`${agentId}_userDepositCount`, this.id);

        return true;
    }


}

module.exports = XCEBikeUser;