/**
 * 用户订单-行程管理
 */

const Sequelize = require('sequelize');


const CREATE_RULE = {
    freezeTableName: true
}

const XCEBikeUser = require('./XCEBikeUser');
const XCEBikeDevice = require('./XCEBikeDevice');
const moment = require('moment');


/**
 * 行程-订单表结构定义
 */

const XCEBikeUserOrder_DS = {
    orderId: {
        type: Sequelize.INTEGER(16).UNSIGNED,
        primaryKey: true,
        autoIncrement: true
    },
    cost: {
        type: Sequelize.INTEGER.UNSIGNED,
    },
    isPaid: {
        type: Sequelize.BOOLEAN,
        defaultValue: false,
        allowNull: false,
    },
    imei: {
        type: Sequelize.STRING(16),
        allowNull: false
    },
    deviceItineraryId: {
        type: Sequelize.INTEGER(16).UNSIGNED,
        allowNull: false
    },
    paidInfo: {
        type: Sequelize.STRING(1024)
    }

};


let xcSequelize = null;
let xcRedis = null;


class XCEBikeUserOrder {

    static log() {
        console.log(xcRedis, xcSequelize);
    }

    /**
     * 配置redis client
     */
    static configRedis(redisClient) {
        xcRedis = redisClient;
    }

    /**
     * 配置数据库实例
     */
    static configSequelize(sequelizeInstance) {
        xcSequelize = sequelizeInstance;
    }


    constructor(userId) {
        this.userId = userId;

        let json = {};
        //订单列表
        this.orderList = json.orderList;
    }


    async getUsrOrderDAO() {
        let agentId = await XCEBikeUser.agentIdByUsrId(this.userId);
        return xcSequelize.define(`xc_ebike_usr_orders_${agentId}_${this.userId}`, XCEBikeUserOrder_DS, CREATE_RULE);
    }


    /**
     * 总订单数+1
     * 同时今日订单数+1
     */
    static async _incrOrderCount(agentId) {
        //总订单数
        let ret = await xcRedis.incr(`${agentId}_orderCount`);
        //今日订单数
        ret = await XCEBikeUserOrder._incrOrderCountToday(agentId);
        return ret;
    }

    /**
     * 今日订单数+1
     * @param {*} agentId 
     * @param {*} fen 
     */
    static async _incrOrderCountToday(agentId) {
        let now = new Date();
        let today = `${now.getFullYear()}.${now.getMonth()+1}.${now.getDate()}`;
        let ret = await xcRedis.incr(`${agentId}_orderCount_${today}`);
        return ret;
    }

    /**
     * 返回代理商总订单数量
     * @param {*} agentId 
     */
    static async orderCount(agentId) {
        let ret = await xcRedis.get(`${agentId}_orderCount`);
        return ret;
    }

    /**
     * 返回代理商今日订单数量
     * @param {*} agentId 
     */
    static async orderCountToday(agentId) {
        let now = new Date();
        let ret = await XCEBikeUserOrder.orderCountSomeday(agentId, now);
        return ret;
    }

    /**
     * 返回代理商某日订单数量
     * @param {*} agentId 
     * @param {*} date  new Date()
     */
    static async orderCountSomeday(agentId, date) {
        let someday = `${date.getFullYear()}.${date.getMonth()+1}.${date.getDate()}`;
        let ret = await xcRedis.get(`${agentId}_orderCount_${someday}`);
        return ret;
    }


    /**
     * 更新流水
     * @param {*} agentId 代理商id
     * @param {*} fen 订单钱数，单位分
     */
    static async _incrIncome(agentId, fen) {
        //总流水
        let ret = await xcRedis.incrby(`${agentId}_incomeSum`, fen);
        //当日流水
        ret = await XCEBikeUserOrder._incrIncomeToday(agentId, fen);
        return ret;
    }

    /**
     * 更新当日流水
     * @param {*} agentId 
     * @param {*} fen 
     */
    static async _incrIncomeToday(agentId, fen) {
        let now = new Date();
        let today = `${now.getFullYear()}.${now.getMonth()+1}.${now.getDate()}`;
        let ret = await xcRedis.incrby(`${agentId}_incomeSum_${today}`, fen);
        return ret;
    }

    /**
     * 代理商总流水，单位:分
     * @param {*} agentId 
     */
    static async income(agentId) {
        let ret = await xcRedis.get(`${agentId}_incomeSum`);
        return ret;
    }

    /**
     * 代理商今日流水，单位：分
     */
    static async incomeToday(agentId) {
        let now = new Date();
        return await XCEBikeUserOrder.incomeSomeday(agentId, now);
    }

    /**
     * 代理商某日流水
     * @param {*} agentId 
     * @param {*} date   new Date()
     */
    static async incomeSomeday(agentId, date) {
        let someday = `${date.getFullYear()}.${date.getMonth() + 1}.${date.getDate()}`;
        let ret = await xcRedis.get(`${agentId}_incomeSum_${someday}`);
        return ret;
    }

    /**
     * 获取用户的订单列表
     * 没有订单返回null
     */
    async readDBOrderList() {
        let userOrderDAO = await this.getUsrOrderDAO();
        let ret = await userOrderDAO.findAll();

        this.orderList = (ret || []).map(orderDAO => {
            return orderDAO.dataValues;
        });

        return this.orderList;
    }

    /**
     * 获取用户的未支付订单信息
     * 没有订单返回null
     */
    async getUnpaidCharge() {
        let userOrderDAO = await this.getUsrOrderDAO();

        let ret = await userOrderDAO.findAll({
            where: {
                isPaid: false
            }
        });

        //todo ret的数据处理

        return ret;
    }

    // /**
    //  * 设置行程订单支付信息
    //  * @param {*} paidOrNot 
    //  * @param {*} orderInfo 
    //  */
    // async setOrderInfo(paidOrNot, orderInfo) {
    //     let orderList = await this.getUnpaidCharge();
    //     let order = orderList[0];
    //     if (paidOrNot) {
    //         order.isPaid = true;
    //         order.paidInfo = orderInfo;

            
    //     } else {
    //         order.isPaid = false;
    //         order.paidInfo = orderInfo;
    //     }

    //     await order.save();
    // }

    /**
     * 生成ping++订单后存订单信息
     * @param {*} chargeInfo 
     */
    async setChargeInfo(chargeInfo) {
        let orderList = await this.getUnpaidCharge();
        if (orderList.length === 0) {
            return {suc: false, info: "无未支付订单"};
        }
        let order = orderList[0];
        order.paidInfo = chargeInfo;
        await order.save()

        return {suc: true};
    }

    /**
     * ping++支付完成后设置订单状态
     * @param {*} chargeInfo 
     */
    async completeCharge(chargeInfo) {
        let orderList = await this.getUnpaidCharge();
        if (orderList.length === 0) {
            return {suc: false, info: "无未支付订单"};
        }
        let order = orderList[0];
        order.paidInfo = chargeInfo;
        order.isPaid = true;
        await order.save()

        //at the end, 更新流水数据
        let usr = new XCEBikeUser(this.userId);
        let agentId = await usr.agentId();
        let cost = order.cost;
        let ret = await XCEBikeUserOrder._incrIncome(agentId, cost);

        return {suc: true};
    }

    /**
     * 生成订单失败，返回false
     * 生成一条用户行程订单。1.设备轨迹生成。2.用户订单生成
     * 依赖user.ridingDeviceIMEI，故流程上要先生成订单，再改变用户和车的相关状态
     */
    async produceOrder() {
        let user = new XCEBikeUser(this.userId);
        let imei = await user.ridingDeviceIMEI();
        let device = await XCEBikeDevice.instanceFromRedis(imei);
        if(device){
            //生成设备行程轨迹
            let itinerary = await device.produceItinerary();
            //生成用户订单
            let order = await this.insertOrderInDB(itinerary);
            
            let unpaidCharge = {
                bikeId: carId,
                isPaid: order.isPaid,
                cost: order.cost,
                startTime: moment(itinerary.startTime).utcOffset(+8).format('YYYY-MM-DD HH:mm:ss'),
                endTime: moment(itinerary.endTime).utcOffset(+8).format('YYYY-MM-DD HH:mm:ss'),
                startLat: itinerary.startLat,
                startLng: itinerary.startLng,
                endLat: itinerary.endLat,
                endLng: itinerary.endLng,
                itinerary: itinerary.itinerary
            };
            
            return unpaidCharge;
        }else{
            return false;
        }
        
    }

    /**
     * 向数据库插入一条用户订单
     */
    async insertOrderInDB(itineraryInfo){
        //根据redis中设备行程信息，生成一条用户订单
        let userOrder = await this.getUsrOrderDAO();

        await userOrder.sync();
        let newOrder = userOrder.create({
            cost: itineraryInfo.cost,
            imei: itineraryInfo.imei,
            deviceItineraryId: itineraryInfo.itinId
        });

        //at the end。更新订单数据
        let usr = new XCEBikeUser(this.userId);
        let agentId = await usr.agentId();
        let ret = await XCEBikeUserOrder._incrOrderCount(agentId);

        return newOrder;
    }


}


module.exports = XCEBikeUserOrder;